@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2009, Sun Microsystems, Inc.
+ * Copyright (c) 2012-2017 Red Hat, Inc. and/or its affiliates.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -56,7 +57,7 @@
  */
 #define XDR_FALSE ((long) 0)
 #define XDR_TRUE ((long) 1)
-#define LASTUNSIGNED ((u_int) 0-1)
+#define RPC_MAXDATASIZE 9000
 
 /*
  * Free a data structure using XDR
@@ -179,19 +180,13 @@ inline_xdr_u_long(XDR *xdrs, u_long *ulp)
 static inline bool
 inline_xdr_int32_t(XDR *xdrs, int32_t *int32_p)
 {
-	long l;
-
 	switch (xdrs->x_op) {
 
 	case XDR_ENCODE:
-		l = (long)*int32_p;
-		return (XDR_PUTLONG(xdrs, &l));
+		return (XDR_PUTINT32(xdrs, *int32_p));
 
 	case XDR_DECODE:
-		if (!XDR_GETLONG(xdrs, &l))
-			return (false);
-		*int32_p = (int32_t) l;
-		return (true);
+		return (XDR_GETINT32(xdrs, int32_p));
 
 	case XDR_FREE:
 		return (true);
@@ -207,19 +202,13 @@ inline_xdr_int32_t(XDR *xdrs, int32_t *int32_p)
 static inline bool
 inline_xdr_u_int32_t(XDR *xdrs, u_int32_t *u_int32_p)
 {
-	u_long l;
-
 	switch (xdrs->x_op) {
 
 	case XDR_ENCODE:
-		l = (u_long) *u_int32_p;
-		return (XDR_PUTLONG(xdrs, (long *)&l));
+		return (XDR_PUTUINT32(xdrs, *u_int32_p));
 
 	case XDR_DECODE:
-		if (!XDR_GETLONG(xdrs, (long *)&l))
-			return (false);
-		*u_int32_p = (u_int32_t) l;
-		return (true);
+		return (XDR_GETUINT32(xdrs, u_int32_p));
 
 	case XDR_FREE:
 		return (true);
@@ -288,19 +277,13 @@ inline_xdr_u_short(XDR *xdrs, u_short *usp)
 static inline bool
 inline_xdr_int16_t(XDR *xdrs, int16_t *int16_p)
 {
-	long l;
-
 	switch (xdrs->x_op) {
 
 	case XDR_ENCODE:
-		l = (long)*int16_p;
-		return (XDR_PUTLONG(xdrs, &l));
+		return (XDR_PUTINT16(xdrs, *int16_p));
 
 	case XDR_DECODE:
-		if (!XDR_GETLONG(xdrs, &l))
-			return (false);
-		*int16_p = (int16_t) l;
-		return (true);
+		return (XDR_GETINT16(xdrs, int16_p));
 
 	case XDR_FREE:
 		return (true);
@@ -315,19 +298,13 @@ inline_xdr_int16_t(XDR *xdrs, int16_t *int16_p)
 static inline bool
 inline_xdr_u_int16_t(XDR *xdrs, u_int16_t *u_int16_p)
 {
-	u_long l;
-
 	switch (xdrs->x_op) {
 
 	case XDR_ENCODE:
-		l = (u_long) *u_int16_p;
-		return (XDR_PUTLONG(xdrs, (long *)&l));
+		return (XDR_PUTUINT16(xdrs, *u_int16_p));
 
 	case XDR_DECODE:
-		if (!XDR_GETLONG(xdrs, (long *)&l))
-			return (false);
-		*u_int16_p = (u_int16_t) l;
-		return (true);
+		return (XDR_GETUINT16(xdrs, u_int16_p));
 
 	case XDR_FREE:
 		return (true);
@@ -421,7 +398,7 @@ inline_xdr_enum(XDR *xdrs, enum_t *ep)
  * cp points to the opaque object and cnt gives the byte length.
  */
 static inline bool
-inline_xdr_getopaque(XDR *xdrs, caddr_t cp, u_int cnt)
+xdr_opaque_decode(XDR *xdrs, char *cp, u_int cnt)
 {
 	u_int rndup;
 
@@ -468,7 +445,7 @@ inline_xdr_getopaque(XDR *xdrs, caddr_t cp, u_int cnt)
  * cp points to the opaque object and cnt gives the byte length.
  */
 static inline bool
-inline_xdr_putopaque(XDR *xdrs, caddr_t cp, u_int cnt)
+xdr_opaque_encode(XDR *xdrs, const char *cp, u_int cnt)
 {
 	u_int rndup;
 
@@ -515,16 +492,16 @@ inline_xdr_putopaque(XDR *xdrs, caddr_t cp, u_int cnt)
  * cp points to the opaque object and cnt gives the byte length.
  */
 static inline bool
-inline_xdr_opaque(XDR *xdrs, caddr_t cp, u_int cnt)
+xdr_opaque(XDR *xdrs, caddr_t cp, u_int cnt)
 {
-	if (xdrs->x_op == XDR_DECODE)
-		return (inline_xdr_getopaque(xdrs, cp, cnt));
-
-	if (xdrs->x_op == XDR_ENCODE)
-		return (inline_xdr_putopaque(xdrs, cp, cnt));
-
-	if (xdrs->x_op == XDR_FREE)
+	switch (xdrs->x_op) {
+	case XDR_DECODE:
+		return (xdr_opaque_decode(xdrs, cp, cnt));
+	case XDR_ENCODE:
+		return (xdr_opaque_encode(xdrs, cp, cnt));
+	case XDR_FREE:
 		return (true);
+	}
 
 	__warnx(TIRPC_DEBUG_FLAG_ERROR,
 		"%s:%u ERROR xdrs->x_op (%u)",
@@ -534,15 +511,12 @@ inline_xdr_opaque(XDR *xdrs, caddr_t cp, u_int cnt)
 }
 
 /*
- * XDR counted bytes
- * *cpp is a pointer to the bytes, *sizep is the count.
- * If *cpp is NULL maxsize bytes are allocated
+ * XDR counted opaques
+ * *sp is a pointer to the bytes, *sizep is the count.
  */
 static inline bool
-inline_xdr_bytes(XDR *xdrs, char **cpp, u_int *sizep,
-		 u_int maxsize)
+xdr_opaques(XDR *xdrs, char *sp, u_int *sizep, u_int maxsize)
 {
-	char *sp = *cpp;	/* sp is the actual string pointer */
 	u_int nodesize;
 
 	/*
@@ -550,6 +524,7 @@ inline_xdr_bytes(XDR *xdrs, char **cpp, u_int *sizep,
 	 */
 	if (!inline_xdr_u_int(xdrs, sizep))
 		return (false);
+
 	nodesize = *sizep;
 	if ((nodesize > maxsize) && (xdrs->x_op != XDR_FREE))
 		return (false);
@@ -558,22 +533,11 @@ inline_xdr_bytes(XDR *xdrs, char **cpp, u_int *sizep,
 	 * now deal with the actual bytes
 	 */
 	switch (xdrs->x_op) {
-
 	case XDR_DECODE:
-		if (nodesize == 0)
-			return (true);
-		if (sp == NULL)
-			*cpp = sp = (char *)mem_alloc(nodesize);
-		return (inline_xdr_getopaque(xdrs, sp, nodesize));
-
+		return (xdr_opaque_decode(xdrs, sp, nodesize));
 	case XDR_ENCODE:
-		return (inline_xdr_putopaque(xdrs, sp, nodesize));
-
+		return (xdr_opaque_encode(xdrs, sp, nodesize));
 	case XDR_FREE:
-		if (sp != NULL) {
-			mem_free(sp, -1);
-			*cpp = NULL;
-		}
 		return (true);
 	}
 	/* NOTREACHED */
@@ -581,16 +545,126 @@ inline_xdr_bytes(XDR *xdrs, char **cpp, u_int *sizep,
 }
 
 /*
- * Implemented here due to commonality of the object.
+ * XDR counted bytes
+ * *cpp is a pointer to the bytes, *sizep is the count.
+ * If *cpp is NULL maxsize bytes are allocated
  */
 static inline bool
-inline_xdr_netobj(XDR *xdrs, struct netobj *np)
+xdr_bytes_decode(XDR *xdrs, char **cpp, u_int *sizep, u_int maxsize)
 {
+	char *sp = *cpp;	/* sp is the actual string pointer */
+	long size;
+	bool ret;
 
-	return (inline_xdr_bytes
-		(xdrs, &np->n_bytes, &np->n_len, MAX_NETOBJ_SZ));
+	/*
+	 * first deal with the length since xdr bytes are counted
+	 */
+	if (!XDR_GETLONG(xdrs, &size)) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s:%u ERROR size",
+			__func__, __LINE__);
+		return (false);
+	}
+	if (size > maxsize) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s:%u ERROR size %ul > max %u",
+			__func__, __LINE__,
+			size, maxsize);
+		return (false);
+	}
+	*sizep = (u_int)size;		/* only valid size */
+
+	/*
+	 * now deal with the actual bytes
+	 */
+	if (!size)
+		return (true);
+	if (!sp)
+		sp = (char *)mem_alloc(size);
+
+	ret = xdr_opaque_decode(xdrs, sp, size);
+	if (!ret) {
+		if (!*cpp) {
+			/* Only free if we allocated */
+			mem_free(sp, size);
+		}
+		return (ret);
+	}
+	*cpp = sp;			/* only valid pointer */
+	return (ret);
 }
 
+static inline bool
+xdr_bytes_encode(XDR *xdrs, char **cpp, u_int *sizep, u_int maxsize)
+{
+	char *sp = *cpp;	/* sp is the actual string pointer */
+	long size = *sizep;
+	u_int nodesize;
+
+	if (size > maxsize) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s:%u ERROR size %ul > max %u",
+			__func__, __LINE__,
+			size, maxsize);
+		return (false);
+	}
+
+	nodesize = (uint32_t)size;
+	if (nodesize < size) {
+		/* caller provided very large maxsize */
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s:%u ERROR overflow %ul",
+			__func__, __LINE__,
+			size);
+		return (false);
+	}
+
+	if (!XDR_PUTLONG(xdrs, &size)) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s:%u ERROR size",
+			__func__, __LINE__);
+		return (false);
+	}
+
+	return (xdr_opaque_encode(xdrs, sp, nodesize));
+}
+
+static inline bool
+xdr_bytes_free(XDR *xdrs, char **cpp, size_t size)
+{
+	if (*cpp) {
+		mem_free(*cpp, size);
+		*cpp = NULL;
+		return (true);
+	}
+
+	/* normal for switch x_op, sometimes useful to track */
+	__warnx(TIRPC_DEBUG_FLAG_XDR,
+		"%s:%u already free",
+		__func__, __LINE__);
+	return (true);
+}
+
+static inline bool
+xdr_bytes(XDR *xdrs, char **cpp, u_int *sizep, u_int maxsize)
+{
+	switch (xdrs->x_op) {
+	case XDR_DECODE:
+		return (xdr_bytes_decode(xdrs, cpp, sizep, maxsize));
+	case XDR_ENCODE:
+		return (xdr_bytes_encode(xdrs, cpp, sizep, maxsize));
+	case XDR_FREE:
+		return (xdr_bytes_free(xdrs, cpp, *sizep));
+	}
+
+	__warnx(TIRPC_DEBUG_FLAG_ERROR,
+		"%s:%u ERROR xdrs->x_op (%u)",
+		__func__, __LINE__,
+		xdrs->x_op);
+	return (false);
+}
+#define inline_xdr_bytes xdr_bytes
+
 /*
  * XDR a descriminated union
  * Support routine for discriminated unions.
@@ -648,73 +722,134 @@ inline_xdr_union(XDR *xdrs, enum_t *dscmp,
  * of the string as specified by a protocol.
  */
 static inline bool
-inline_xdr_string(XDR *xdrs, char **cpp, u_int maxsize)
+xdr_string_decode(XDR *xdrs, char **cpp, u_int maxsize)
 {
 	char *sp = *cpp;	/* sp is the actual string pointer */
-	u_int size = 0;		/* XXX remove warning */
+	long size;
 	u_int nodesize;
+	bool ret;
 
 	/*
 	 * first deal with the length since xdr strings are counted-strings
 	 */
-	switch (xdrs->x_op) {
-	case XDR_FREE:
-		if (sp == NULL)
-			return (true);	/* already free */
-		break;
-	case XDR_ENCODE:
-		if (sp == NULL)
-			return false;
-		size = strlen(sp);
-		break;
-	case XDR_DECODE:
-		break;
+	if (!XDR_GETLONG(xdrs, &size)) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s:%u ERROR size",
+			__func__, __LINE__);
+		return (false);
 	}
-	if (!inline_xdr_u_int(xdrs, &size))
+
+	if (size > maxsize) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s:%u ERROR size %ul > max %u",
+			__func__, __LINE__,
+			size, maxsize);
 		return (false);
-	if (size > maxsize && xdrs->x_op != XDR_FREE)
+	}
+
+	nodesize = (uint32_t)(size + 1);
+	if (nodesize < (size + 1)) {
+		/* caller provided very large maxsize */
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s:%u ERROR overflow %ul",
+			__func__, __LINE__,
+			size);
 		return (false);
-	nodesize = size + 1;
-	if (nodesize == 0) {
-		/* This means an overflow.  It a bug in the caller which
-		 * provided a too large maxsize but nevertheless catch it
-		 * here.
-		 */
-		return false;
 	}
 
 	/*
 	 * now deal with the actual bytes
 	 */
-	switch (xdrs->x_op) {
+	if (!sp)
+		sp = (char *)mem_alloc(nodesize);
 
-	case XDR_DECODE:
-		if (sp == NULL)
-			*cpp = sp = (char *)mem_alloc(nodesize);
-		sp[size] = 0;
-		return (inline_xdr_getopaque(xdrs, sp, size));
+	ret = xdr_opaque_decode(xdrs, sp, size);
+	if (!ret) {
+		mem_free(sp, nodesize);
+		return (ret);
+	}
+	sp[size] = '\0';
+	*cpp = sp;			/* only valid pointer */
+	return (ret);
+}
 
-	case XDR_ENCODE:
-		return (inline_xdr_putopaque(xdrs, sp, size));
+static inline bool
+xdr_string_encode(XDR *xdrs, char **cpp, u_int maxsize)
+{
+	long size;
+	u_int nodesize;
 
-	case XDR_FREE:
-		mem_free(sp, -1);
+	if (!(*cpp)) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s:%u ERROR missing string pointer",
+			__func__, __LINE__);
+		return (false);
+	}
+
+	size = strlen(*cpp);
+	if (size > maxsize) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s:%u ERROR size %ul > max %u",
+			__func__, __LINE__,
+			size, maxsize);
+		return (false);
+	}
+
+	nodesize = (uint32_t)(size + 1);
+	if (nodesize < (size + 1)) {
+		/* caller provided very large maxsize */
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s:%u ERROR overflow %ul",
+			__func__, __LINE__,
+			size);
+		return (false);
+	}
+
+	if (!XDR_PUTLONG(xdrs, &size)) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s:%u ERROR size",
+			__func__, __LINE__);
+		return (false);
+	}
+
+	return (xdr_opaque_encode(xdrs, *cpp, size));
+}
+
+static inline bool
+xdr_string_free(XDR *xdrs, char **cpp)
+{
+	if (*cpp) {
+		mem_free(*cpp, strlen(*cpp) + 1);
 		*cpp = NULL;
 		return (true);
 	}
-	/* NOTREACHED */
-	return (false);
+
+	/* normal for switch x_op, sometimes useful to track */
+	__warnx(TIRPC_DEBUG_FLAG_XDR,
+		"%s:%u already free",
+		__func__, __LINE__);
+	return (true);
 }
 
-/*
- * Wrapper for xdr_string that can be called directly from
- * routines like clnt_call
- */
 static inline bool
-inline_xdr_wrapstring(XDR *xdrs, char **cpp)
+xdr_string(XDR *xdrs, char **cpp, u_int maxsize)
 {
-	return inline_xdr_string(xdrs, cpp, LASTUNSIGNED);
+	switch (xdrs->x_op) {
+	case XDR_DECODE:
+		return (xdr_string_decode(xdrs, cpp, maxsize));
+	case XDR_ENCODE:
+		return (xdr_string_encode(xdrs, cpp, maxsize));
+	case XDR_FREE:
+		return (xdr_string_free(xdrs, cpp));
+	}
+
+	__warnx(TIRPC_DEBUG_FLAG_ERROR,
+		"%s:%u ERROR xdrs->x_op (%u)",
+		__func__, __LINE__,
+		xdrs->x_op);
+	return (false);
 }
+#define inline_xdr_string xdr_string
 
 /*
  * NOTE: xdr_hyper(), xdr_u_hyper(), xdr_longlong_t(), and xdr_u_longlong_t()
