@@ -60,9 +60,6 @@
 #include <rpc/types.h>
 #include <misc/portable.h>
 #include <rpc/rpc.h>
-#ifdef PORTMAP
-#include <rpc/pmap_clnt.h>
-#endif				/* PORTMAP */
 
 #include "rpc_com.h"
 
@@ -83,18 +80,21 @@
 #define SVC_VERSQUIET 0x0001	/* keep quiet about vers mismatch */
 #define version_keepquiet(xp) ((u_long)(xp)->xp_p3 & SVC_VERSQUIET)
 
+#define SVC_WORK_POOL_THRD_MIN (2)
+
 /* svc_internal.h */
 #ifdef IOV_MAX
 int __svc_maxiov = IOV_MAX;
 #else
 int __svc_maxiov = 1024; /* UIO_MAXIOV value from sys/uio.h */
 #endif
-int __svc_maxrec = 0;
+int __svc_maxrec = RPC_MAXDATA_LEGACY;
 
 struct svc_params __svc_params[1] = {
 	{
-	 false /* !initialized */ ,
-	 MUTEX_INITIALIZER}
+		MUTEX_INITIALIZER,
+		false /* !initialized */ ,
+	}
 };
 
 /*
@@ -119,17 +119,6 @@ static struct svc_callout *svc_find(rpcprog_t, rpcvers_t, struct svc_callout **,
 
 struct work_pool svc_work_pool;
 
-static int
-svc_work_pool_init()
-{
-	struct work_pool_params params = {
-		.thrd_max = __svc_params->ioq.thrd_max,
-		.thrd_min = 2
-	};
-
-	return work_pool_init(&svc_work_pool, "svc_work_pool", &params);
-}
-
 /* Package init function.
  * It is intended that applications which must make use of global state
  * will call svc_init() before accessing such state and before executing
@@ -138,28 +127,28 @@ svc_work_pool_init()
 bool
 svc_init(svc_init_params *params)
 {
+	struct work_pool_params work_pool_params;
+	uint32_t channels = params->channels ? params->channels : 8;
+
 	mutex_lock(&__svc_params->mtx);
 	if (__svc_params->initialized) {
-		__warnx(TIRPC_DEBUG_FLAG_SVC,
+		__warnx(TIRPC_DEBUG_FLAG_WARN,
 			"%s: multiple initialization attempt (nothing happens)",
 			__func__);
 		mutex_unlock(&__svc_params->mtx);
 		return true;
 	}
+	__svc_params->disconnect_cb = params->disconnect_cb;
+	__svc_params->request_cb = params->request_cb;
+
 	__svc_params->max_connections =
 	    (params->max_connections) ? params->max_connections : FD_SETSIZE;
 
-	/* svc_vc */
-	__svc_params->xprt_u.vc.nconns = 0;
-	mutex_init(&__svc_params->xprt_u.vc.mtx, NULL);
-
-	svc_ioq_init();
-
 #if defined(HAVE_BLKIN)
 	if (params->flags & SVC_INIT_BLKIN) {
 		int r = blkin_init();
 		if (r < 0) {
-			__warnx(TIRPC_DEBUG_FLAG_SVC,
+			__warnx(TIRPC_DEBUG_FLAG_ERROR,
 				"%s: blkn_init failed\n",
 				__func__);
 		}
@@ -177,21 +166,39 @@ svc_init(svc_init_params *params)
 #endif
 	__svc_params->idle_timeout = params->idle_timeout;
 
-	/* XXX defaults to the legacy max sendsz */
-	__svc_params->svc_ioq_maxbuf =
-	    (params->svc_ioq_maxbuf) ? (params->svc_ioq_maxbuf) : 262144;
-
 	/* allow consumers to manage all xprt registration */
 	if (params->flags & SVC_INIT_NOREG_XPRTS)
 		__svc_params->flags |= SVC_FLAG_NOREG_XPRTS;
 
-	if (params->ioq_thrd_max)
-		__svc_params->ioq.thrd_max = params->ioq_thrd_max;
+	if (params->ioq_send_max)
+		__svc_params->ioq.send_max = params->ioq_send_max;
 	else
-		__svc_params->ioq.thrd_max = 200;
+		__svc_params->ioq.send_max = RPC_MAXDATA_DEFAULT;
 
-	/* uses ioq.thrd_max */
-	if (svc_work_pool_init()) {
+	__svc_params->ioq.thrd_min = SVC_WORK_POOL_THRD_MIN;
+	if (__svc_params->ioq.thrd_min < params->ioq_thrd_min)
+		__svc_params->ioq.thrd_min = params->ioq_thrd_min;
+
+	__svc_params->ioq.thrd_max = __svc_params->ioq.thrd_min;
+	if (__svc_params->ioq.thrd_max < params->ioq_thrd_max)
+		__svc_params->ioq.thrd_max = params->ioq_thrd_max;
+
+	svc_ioq_init();
+
+	work_pool_params.thrd_min = __svc_params->ioq.thrd_min + channels;
+	work_pool_params.thrd_max = __svc_params->ioq.thrd_max;
+	if (work_pool_params.thrd_max < work_pool_params.thrd_min)
+		work_pool_params.thrd_max = work_pool_params.thrd_min;
+
+	if (work_pool_init(&svc_work_pool, "svc_", &work_pool_params)) {
+		mutex_unlock(&__svc_params->mtx);
+		return false;
+	}
+
+	/* uses svc_work_pool */
+	svc_rqst_init(channels);
+
+	if (svc_xprt_init()) {
 		mutex_unlock(&__svc_params->mtx);
 		return false;
 	}
@@ -227,7 +234,15 @@ svc_init(svc_init_params *params)
 	mutex_unlock(&__svc_params->mtx);
 
 #if defined(_SC_IOV_MAX) /* IRIX, MacOS X, FreeBSD, Solaris, ... */
-	__svc_maxiov = sysconf(_SC_IOV_MAX);
+	{
+		/*
+		 * some glibc (e.g. 2.26 in Fedora 27 beta) always
+		 * return -1
+		 */
+		int i = sysconf(_SC_IOV_MAX);
+		if (i != -1 && i > __svc_maxiov)
+			__svc_maxiov = i;
+	}
 #endif
 	return true;
 }
@@ -384,68 +399,6 @@ svc_unreg(const rpcprog_t prog, const rpcvers_t vers)
 
 /* ********************** CALLOUT list related stuff ************* */
 
-#ifdef PORTMAP
-/*
- * Add a service program to the callout list.
- * The dispatch routine will be called when a rpc request for this
- * program number comes in.
- */
-bool
-svc_register(SVCXPRT *xprt, u_long prog, u_long vers,
-	     void (*dispatch) (struct svc_req *req),
-	     int protocol)
-{
-	struct svc_callout *prev;
-	struct svc_callout *s;
-
-	assert(xprt != NULL);
-	assert(dispatch != NULL);
-
-	s = svc_find((rpcprog_t) prog, (rpcvers_t) vers, &prev, NULL);
-	if (s) {
-		if (s->rec.sc_dispatch == dispatch)
-			goto pmap_it;	/* he is registering another xprt */
-		return (false);
-	}
-	s = mem_alloc(sizeof(struct svc_callout));
-	s->rec.sc_prog = (rpcprog_t) prog;
-	s->rec.sc_vers = (rpcvers_t) vers;
-	s->rec.sc_dispatch = dispatch;
-	s->sc_next = svc_head;
-	svc_head = s;
-
- pmap_it:
-	/* now register the information with the local binder service */
-	if (protocol)
-		return (pmap_set(prog, vers, protocol,
-				 __rpc_address_port(&xprt->xp_local)));
-
-	return (true);
-}
-
-/*
- * Remove a service program from the callout list.
- */
-void
-svc_unregister(u_long prog, u_long vers)
-{
-	struct svc_callout *prev;
-	struct svc_callout *s;
-
-	s = svc_find((rpcprog_t) prog, (rpcvers_t) vers, &prev, NULL);
-	if (!s)
-		return;
-	if (prev == NULL)
-		svc_head = s->sc_next;
-	else
-		prev->sc_next = s->sc_next;
-	s->sc_next = NULL;
-	mem_free(s, sizeof(struct svc_callout));
-	/* now unregister the information with the local binder service */
-	(void)pmap_unset(prog, vers);
-}
-#endif				/* PORTMAP */
-
 /*
  * Search the callout list for a program number, return the callout
  * struct.
@@ -471,7 +424,7 @@ svc_find(rpcprog_t prog, rpcvers_t vers,
 }
 
 /* An exported search routing similar to svc_find, but with error reporting
- * needed by svc_getreq routines. */
+ */
 svc_lookup_result_t
 svc_lookup(svc_rec_t **rec, svc_vers_range_t *vrange,
 	   rpcprog_t prog, rpcvers_t vers, char *netid,
@@ -537,23 +490,22 @@ svc_lookup(svc_rec_t **rec, svc_vers_range_t *vrange,
  * Send a reply to an rpc request (MT-SAFE).
  *
  */
-bool
-svc_sendreply(struct svc_req *req, xdrproc_t xdr_results, void *xdr_location)
+enum xprt_stat
+svc_sendreply(struct svc_req *req)
 {
 	assert(req != NULL);
+	assert(req->rq_msg.RPCM_ack.ar_results.proc != NULL);
 
 	req->rq_msg.rm_direction = REPLY;
 	req->rq_msg.rm_reply.rp_stat = MSG_ACCEPTED;
 	req->rq_msg.RPCM_ack.ar_stat = SUCCESS;
-	req->rq_msg.RPCM_ack.ar_results.where = xdr_location;
-	req->rq_msg.RPCM_ack.ar_results.proc = xdr_results;
-	return (SVC_REPLY(req));
+	return SVC_REPLY(req);
 }
 
 /*
  * No procedure error reply (MT-SAFE)
  */
-void
+enum xprt_stat
 svcerr_noproc(struct svc_req *req)
 {
 	assert(req != NULL);
@@ -561,13 +513,13 @@ svcerr_noproc(struct svc_req *req)
 	req->rq_msg.rm_direction = REPLY;
 	req->rq_msg.rm_reply.rp_stat = MSG_ACCEPTED;
 	req->rq_msg.RPCM_ack.ar_stat = PROC_UNAVAIL;
-	SVC_REPLY(req);
+	return SVC_REPLY(req);
 }
 
 /*
  * Can't decode args error reply (MT-SAFE)
  */
-void
+enum xprt_stat
 svcerr_decode(struct svc_req *req)
 {
 	assert(req != NULL);
@@ -575,13 +527,13 @@ svcerr_decode(struct svc_req *req)
 	req->rq_msg.rm_direction = REPLY;
 	req->rq_msg.rm_reply.rp_stat = MSG_ACCEPTED;
 	req->rq_msg.RPCM_ack.ar_stat = GARBAGE_ARGS;
-	SVC_REPLY(req);
+	return SVC_REPLY(req);
 }
 
 /*
  * Some system error (MT-SAFE)
  */
-void
+enum xprt_stat
 svcerr_systemerr(struct svc_req *req)
 {
 	assert(req != NULL);
@@ -589,7 +541,7 @@ svcerr_systemerr(struct svc_req *req)
 	req->rq_msg.rm_direction = REPLY;
 	req->rq_msg.rm_reply.rp_stat = MSG_ACCEPTED;
 	req->rq_msg.RPCM_ack.ar_stat = SYSTEM_ERR;
-	SVC_REPLY(req);
+	return SVC_REPLY(req);
 }
 
 #if 0
@@ -633,7 +585,7 @@ __svc_versquiet_get(SVCXPRT *xprt)
 /*
  * Authentication error reply (MT-SAFE)
  */
-void
+enum xprt_stat
 svcerr_auth(struct svc_req *req, enum auth_stat why)
 {
 	assert(req != NULL);
@@ -642,22 +594,22 @@ svcerr_auth(struct svc_req *req, enum auth_stat why)
 	req->rq_msg.rm_reply.rp_stat = MSG_DENIED;
 	req->rq_msg.RPCM_rej.rj_stat = AUTH_ERROR;
 	req->rq_msg.RPCM_rej.rj_why = why;
-	SVC_REPLY(req);
+	return SVC_REPLY(req);
 }
 
 /*
  * Auth too weak error reply (MT-SAFE)
  */
-void
+enum xprt_stat
 svcerr_weakauth(struct svc_req *req)
 {
-	svcerr_auth(req, AUTH_TOOWEAK);
+	return svcerr_auth(req, AUTH_TOOWEAK);
 }
 
 /*
  * Program unavailable error reply (MT-SAFE)
  */
-void
+enum xprt_stat
 svcerr_noprog(struct svc_req *req)
 {
 	assert(req != NULL);
@@ -665,13 +617,13 @@ svcerr_noprog(struct svc_req *req)
 	req->rq_msg.rm_direction = REPLY;
 	req->rq_msg.rm_reply.rp_stat = MSG_ACCEPTED;
 	req->rq_msg.RPCM_ack.ar_stat = PROG_UNAVAIL;
-	SVC_REPLY(req);
+	return SVC_REPLY(req);
 }
 
 /*
  * Program version mismatch error reply
  */
-void
+enum xprt_stat
 svcerr_progvers(struct svc_req *req, rpcvers_t low_vers, rpcvers_t high_vers)
 {
 	assert(req != NULL);
@@ -681,109 +633,11 @@ svcerr_progvers(struct svc_req *req, rpcvers_t low_vers, rpcvers_t high_vers)
 	req->rq_msg.RPCM_ack.ar_stat = PROG_MISMATCH;
 	req->rq_msg.RPCM_ack.ar_vers.low = (u_int32_t) low_vers;
 	req->rq_msg.RPCM_ack.ar_vers.high = (u_int32_t) high_vers;
-	SVC_REPLY(req);
+	return SVC_REPLY(req);
 }
 
 /* ******************* SERVER INPUT STUFF ******************* */
 
-/*
- * Get server side input from some transport.
- */
-
-#if !defined(_WIN32)		/* XXX */
-void
-svc_getreq(int rdfds)
-{
-	fd_set readfds;
-
-	FD_ZERO(&readfds);
-	readfds.fds_bits[0] = rdfds;
-	svc_getreqset(&readfds);
-}
-
-void
-svc_getreqset(fd_set *readfds)
-{
-	int bit, fd;
-	fd_mask mask, *maskp;
-	int sock;
-
-	assert(readfds != NULL);
-
-	maskp = readfds->fds_bits;
-	for (sock = 0; sock < FD_SETSIZE; sock += NFDBITS) {
-		for (mask = *maskp++; (bit = ffsl(mask)) != 0;
-		     mask ^= (1L << (bit - 1))) {
-			/* sock has input waiting */
-			fd = sock + bit - 1;
-			svc_getreq_common(fd);
-		}
-	}
-}
-
-void
-svc_getreq_poll(struct pollfd *pfdp, int pollretval)
-{
-	int i;
-	int fds_found;
-
-	for (i = fds_found = 0; fds_found < pollretval; i++) {
-		struct pollfd *p = &pfdp[i];
-
-		if (p->revents) {
-			/* fd has input waiting */
-			fds_found++;
-			/*
-			 *      We assume that this function is only called
-			 *      via someone _select()ing from svc_fdset or
-			 *      _poll()ing from svc_pollset[].  Thus it's safe
-			 *      to handle the POLLNVAL event by simply turning
-			 *      the corresponding bit off in svc_fdset.  The
-			 *      svc_pollset[] array is derived from svc_fdset
-			 *      and so will also be updated eventually.
-			 *
-			 *      XXX Should we do an xprt_unregister() instead?
-			 */
-			if (!(p->revents & POLLNVAL))
-				svc_getreq_common(p->fd);
-		}
-	}
-}
-#endif				/* _WIN32 */
-
-#if defined(TIRPC_EPOLL)
-void
-svc_getreqset_epoll(struct epoll_event *events, int nfds)
-{
-	int ix, code __attribute__ ((unused)) = 0;
-	SVCXPRT *xprt;
-
-	assert(events != NULL);
-
-	for (ix = 0; ix < nfds; ++ix) {
-		/* XXX should do constant-time validity check */
-		xprt = (SVCXPRT *) events[ix].data.ptr;
-		code = xprt->xp_ops->xp_getreq(xprt);
-	}
-}
-#endif				/* TIRPC_EPOLL */
-
-void
-svc_getreq_common(int fd)
-{
-	SVCXPRT *xprt;
-	bool code __attribute__ ((unused));
-
-	xprt = svc_xprt_get(fd);
-
-	if (xprt == NULL)
-		return;
-
-	code = xprt->xp_ops->xp_getreq(xprt);
-
-	return;
-}
-
 /* Allow internal or external getreq routines to validate xprt
  * has not been recursively disconnected.  (I don't completely buy the
  * logic, but it should be unchanged (Matt)) */
@@ -792,137 +646,13 @@ svc_validate_xprt_list(SVCXPRT *xprt)
 {
 	bool code;
 
-	code = (xprt == svc_xprt_get(xprt->xp_fd));
+	code = (xprt == svc_xprt_lookup(xprt->xp_fd, NULL));
+	if (code)
+		rpc_dplx_rui(REC_XPRT(xprt));
 
 	return (code);
 }
 
-void
-svc_dispatch_default(SVCXPRT *xprt, struct rpc_msg **ind_msg)
-{
-	struct svc_req r;
-	struct rpc_msg *msg = *ind_msg;
-	svc_vers_range_t vrange;
-	svc_lookup_result_t lkp_res;
-	svc_rec_t *svc_rec;
-	enum auth_stat why;
-	bool no_dispatch = false;
-
-	r.rq_xprt = xprt;
-	r.rq_msg.cb_prog = msg->cb_prog;
-	r.rq_msg.cb_vers = msg->cb_vers;
-	r.rq_msg.cb_proc = msg->cb_proc;
-	r.rq_msg.cb_cred = msg->cb_cred;
-	r.rq_msg.rm_xid = msg->rm_xid;
-
-	/* first authenticate the message */
-	why = svc_auth_authenticate(&r, &no_dispatch);
-	if ((why != AUTH_OK) || no_dispatch) {
-		svcerr_auth(&r, why);
-		return;
-	}
-
-	lkp_res = svc_lookup(&svc_rec, &vrange, r.rq_msg.cb_prog,
-			     r.rq_msg.cb_vers, NULL, 0);
-	switch (lkp_res) {
-	case SVC_LKP_SUCCESS:
-		/* call it */
-		(*svc_rec->sc_dispatch) (&r);
-		return;
-	case SVC_LKP_VERS_NOTFOUND:
-		svcerr_progvers(&r, vrange.lowvers, vrange.highvers);
-		break;
-	default:
-		svcerr_noprog(&r);
-		break;
-	}
-}
-
-bool
-svc_getreq_default(SVCXPRT *xprt)
-{
-	enum xprt_stat stat;
-	struct svc_req req = {.rq_xprt = xprt };
-	bool no_dispatch = false;
-
-	/* XXX !MT-SAFE */
-
-	/* now receive msgs from xprt (support batch calls) */
-	do {
-		if (SVC_RECV(&req)) {
-
-			/* now find the exported program and call it */
-			svc_vers_range_t vrange;
-			svc_lookup_result_t lkp_res;
-			svc_rec_t *svc_rec;
-			enum auth_stat why;
-
-			/* first authenticate the message */
-			why = svc_auth_authenticate(&req, &no_dispatch);
-			if ((why != AUTH_OK) || no_dispatch) {
-				svcerr_auth(&req, why);
-				goto call_done;
-			}
-
-			lkp_res =
-			    svc_lookup(&svc_rec, &vrange, req.rq_msg.cb_prog,
-				       req.rq_msg.cb_vers, NULL, 0);
-			switch (lkp_res) {
-			case SVC_LKP_SUCCESS:
-				(*svc_rec->sc_dispatch) (&req);
-				goto call_done;
-				break;
-			case SVC_LKP_VERS_NOTFOUND:
-				__warnx(TIRPC_DEBUG_FLAG_SVC,
-					"%s: dispatch prog vers notfound\n",
-					__func__);
-				svcerr_progvers(&req, vrange.lowvers,
-						vrange.highvers);
-				break;
-			default:
-				__warnx(TIRPC_DEBUG_FLAG_SVC,
-					"%s: dispatch prog notfound\n",
-					__func__);
-				svcerr_noprog(&req);
-				break;
-			}
-
-		}
-		/* SVC_RECV again? */
- call_done:
-		stat = SVC_STAT(xprt);
-		if (stat == XPRT_DIED) {
-			__warnx(TIRPC_DEBUG_FLAG_SVC,
-				"%s: stat == XPRT_DIED (%p)\n", __func__,
-				xprt);
-			SVC_DESTROY(xprt);
-			SVC_RELEASE(xprt, SVC_RELEASE_FLAG_NONE);
-			xprt = 0;
-			break;
-
-		} else {
-			/*
-			 * Check if the xprt has been disconnected in a
-			 * recursive call in the service dispatch routine.
-			 * If so, then break.
-			 */
-			if (!svc_validate_xprt_list(xprt)) {
-				SVC_RELEASE(xprt, SVC_RELEASE_FLAG_NONE);
-				xprt = 0;
-				break;
-			}
-		}
-
-	} while (stat == XPRT_MOREREQS);
-
-	if (xprt) {
-		svc_rqst_rearm_events(xprt, SVC_RQST_FLAG_NONE);
-		SVC_RELEASE(xprt, SVC_RELEASE_FLAG_NONE);
-	}
-
-	return (stat);
-}
-
 bool
 rpc_control(int what, void *arg)
 {
@@ -989,6 +719,18 @@ void __rpc_set_blkin_endpoint(SVCXPRT *xprt, const char *tag)
 }
 #endif
 
+enum xprt_stat
+svc_rendezvous_stat(SVCXPRT *xprt)
+{
+	if (!xprt)
+		return (XPRT_IDLE);
+
+	if (xprt->xp_flags & SVC_XPRT_FLAG_DESTROYED)
+		return (XPRT_DESTROYED);
+
+	return (XPRT_IDLE);
+}
+
 int
 svc_shutdown(u_long flags)
 {
@@ -999,15 +741,15 @@ svc_shutdown(u_long flags)
 	rpc_rdma_internals_fini();
 #endif
 
-	/* finalize ioq */
-	work_pool_shutdown(&svc_work_pool);
-
 	/* dispose all xprts and support */
 	svc_xprt_shutdown();
 
 	/* release request event channels */
 	svc_rqst_shutdown();
 
+	/* release workers after event channels */
+	work_pool_shutdown(&svc_work_pool);
+
 	/* XXX assert quiescent */
 
 	return (code);
